(* gpp.ml *)
(* GRAPL Pretty Printer. Generates GRAPL code from a grapl-ast. For debugging purposes. *)
(* @author: Lili Chen *)

open Ast

let string_of_operator = function
	| Add -> "+" | Sub -> "-" | Mult -> "*" | Div -> "/"
  | Equal -> "==" | Neq -> "!=" | Less -> "<" | Leq -> "<=" 
	| Greater -> ">" | Geq -> ">=" | Noop -> "" | Cons -> "::"
	| Not -> "!" | Inc -> "++" | Dec -> "--"
	
let rec string_of_expr = function
  | Literal(l) -> string_of_float l
	| BooleanLiteral(s) -> s
  | Id(s) -> s
  | Binop(e1, o, e2) ->
      string_of_expr e1 ^ " " ^ string_of_operator(o) ^ " " ^ string_of_expr e2
  | Assign(v, e) -> v ^ " = " ^ string_of_expr e
	| Call(f, el) ->
      f ^ "(" ^ String.concat ", " (List.map string_of_expr el) ^ ")"
  | Noexpr -> ""
  | Unvisit(e) -> "unvisit " ^ e
  | Visit(e) ->  "visit " ^ e 
	| List(l) -> "[" ^ String.concat "," (List.map string_of_expr l) ^ "]"
	| Unop(Not, e) -> "!" ^ string_of_expr e
	| Unop(op, e) -> string_of_expr e ^ string_of_operator op

let string_of_pred = function
	| (Noop, Literal(1.0)) -> ""
	| (p,e) -> "with (" ^ string_of_operator(p) ^ " " ^ string_of_expr e ^ ")"

let string_of_edge = function
	| Redge(w) -> ">>" ^ string_of_float(w) 
	| Ledge(w) -> "<<" ^ string_of_float(w) 
	| Bedge(w) -> "<>" ^ string_of_float(w) 
	| NoEdge(w) -> ""

let rec string_of_graph = function
	| [(n1, NoEdge(0.0), n2)] -> n1 ^ " "
	| (n1, NoEdge(0.0), n2) :: tail -> n1 ^ ", " ^ string_of_graph(tail)
	| (n1, edge, n2) :: tail -> n1 ^ " " ^ string_of_edge(edge) ^ " " ^ string_of_graph(tail)
	| [] -> ""

let rec string_of_stmt = function
    Block(stmts) ->
      "{\n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "}\n"
  | Expr(expr) -> string_of_expr expr ^ ";\n";
  | IfThenElse(e, s1, s2) ->  "if (" ^ string_of_expr e ^ ")\n" ^
      string_of_stmt s1 ^ "else\n" ^ string_of_stmt s2 ^ "\n"
 	| Foreach(q,n1,ft,n2,(s,e),st) -> "foreach " ^ q ^ " " ^ n1 ^ " " ^ ft ^ " " ^ n2 ^ " " ^ string_of_pred((s,e)) ^ string_of_stmt(st) ^ "\n"
	| Graph(ses_list) -> "graph [" ^ string_of_graph(ses_list) ^ "];\n"
	| Return(e) -> "return " ^ string_of_expr e ^ ";\n"
	| While(e,s) -> "while (" ^ string_of_expr e ^ ")\n" ^ string_of_stmt s
									      
let string_of_vdecl var = fst(var) ^ " " ^ snd(var) ^ ";\n"

let string_of_fdecl f =
	f.rettype ^ " " ^ f.fname ^ "(" ^ String.concat "," (List.map (fun t  -> fst t ^ " " ^ snd t) f.formals) ^ ")\n{\n" ^
		String.concat "\n" (List.map (fun t  -> fst t ^ " " ^ snd t ^ ";") f.locals) ^ "\n" ^ String.concat "" (List.map string_of_stmt f.body) ^ "}\n"

let string_of_program (vars, funcs) =
  String.concat "" (List.map string_of_vdecl vars) ^ "\n" ^
  String.concat "\n" (List.map string_of_fdecl funcs)
	
		 
