/* parser.mly */
/* @authors: Di Wen, Andres Uribe, Ryan Turner */

%{ open Ast  %}

%token LBRACK RBRACK LPAREN RPAREN LBRACE RBRACE 
%token SEMICOLON COMMA 
%token PLUS MINUS TIMES DIVIDE AND OR
%token EQ NEQ LT LEQ GT GEQ NOT ASSIGN CONS INC DEC
%token VOID GRAPH NODE LIST NUMBER BOOLEAN
%token REDGE LEDGE BEDGE HEAD TAIL
%token IF THEN ELSE FOREACH WITH RETURN WHILE
%token FROM TO
%token VISITED ISVISITED UNVISITED VISIT UNVISIT
%token <string> ID
%token EOF
%token <float> LITERAL
%token <string> BOOLEAN_LITERAL

%nonassoc ELSE
%right ASSIGN
%right CONS
%left OR
%left AND
%left EQ NEQ
%left LT GT LEQ GEQ
%left PLUS MINUS
%left TIMES DIVIDE
%right NOT
%left INC
%left DEC

%start program
%type < Ast.program> program

%%

program:
   /* nothing */ { [], [] }
 | program vdecl { ($2 :: fst $1), snd $1 }
 | program fdecl { fst $1, ($2 :: snd $1) }

fdecl:
   types ID LPAREN formals_opt RPAREN LBRACE vdecls stmts RBRACE
     { { rettype = $1; 
       fname = $2;
	 formals = $4; (* a list of tuples *)
	 locals = List.rev $7; (* a list of tuples *)
	 body =  List.rev $8 (* a list of statements *)} }

types: /* note that graph is not a type */
	| NUMBER { "number" }
	| NODE   { "node" }
	| BOOLEAN { "boolean" }
	| LIST { "list" }
	| VOID { "void" }
		    
formals_opt:
    /* nothing */ { [] }
  | formal_list   { List.rev $1 }

formal_list:
    types ID                   { [($1 , $2)] }
  | formal_list COMMA types ID { ($3, $4) :: $1 }

vdecls:
 	/*nothing*/ { [] }
  | vdecls vdecl { $2::$1 }
   
vdecl:
	| types ID SEMICOLON { ($1, $2) }

stmts:
	/* nothing */ { [] }
	| stmts stmt { $2::$1 }

stmt:
	| LBRACE stmts RBRACE { Block(List.rev $2) } 
	| IF LPAREN expr RPAREN THEN stmt ELSE stmt { IfThenElse($3, $6, $8) } 
	/*| IF LPAREN expr RPAREN THEN stmts %prec NOELSE { IfThenElse($3, $6, Block([])) }*/
	| FOREACH qualifiers ID fromTo ID withStmt stmt { Foreach($2, $3, $4, $5, $6, $7)} 
	| expr SEMICOLON { Expr($1) }
	| GRAPH LBRACK consdecls_opt RBRACK SEMICOLON { Graph(List.rev $3)} 
	| RETURN expr SEMICOLON { Return($2) } 
	| WHILE LPAREN expr RPAREN stmt { While($3, $5) }
 
qualifiers:
	| /* nothing */ { "null" }
	| VISITED  { "visited" }
	| UNVISITED{ "unvisited" }

fromTo:
	| FROM { "from" }
	| TO { "to" }
  
withStmt:
	| /*nothing */ {(Noop, Literal(1.0)) }
	| WITH LPAREN pred RPAREN { $3 }

pred: 
	| EQ     expr { (Equal, $2) }
	| NEQ    expr { (Neq, $2) }
	| LT     expr { (Less, $2) }
	| LEQ    expr { (Leq, $2) }
	| GT     expr { (Greater, $2) }
	| GEQ    expr { (Geq, $2) }
    
consdecls_opt: /* a list of triples  */
  | /*nothing*/ { [] }
  | consdecl_list { $1 }
/*| consdecls COMMA consdecl { $3 @ $1 }*/

consdecl_list:
  | consdecl{ List.rev $1 }
  | consdecl_list COMMA consdecl { List.rev $3 @  $1 }

consdecl: /* a list of triples */ 
	| ID edgeDefs 
		{ let rec help a b = match b with
			| [] -> [(a,NoEdge(0.0),a)]
			| head :: tail -> (a, fst head, snd head) :: help (snd head) (tail) in help $1 (List.rev $2) 
		}

edgeDefs: /* a list of tuples*/
	 /*nothing*/ { [] }
	| edgeDefs edgeDef { $2::$1 }

edgeDef: /* a tuple of the form ( <<3 , b) */
	| REDGE ID { (Redge(1.0), $2) } 
	| LEDGE ID { (Ledge(1.0), $2) }
	| BEDGE ID { (Bedge(1.0), $2) }
	| REDGE LITERAL ID { (Redge($2), $3) }
	| LEDGE LITERAL ID { (Ledge($2), $3) }
	| BEDGE LITERAL ID { (Bedge($2), $3) }

expr:
	| ID { Id($1) }
	| expr PLUS  expr { Binop($1, Add, $3) }
	| expr MINUS expr { Binop($1, Sub, $3) }
	| expr TIMES  expr { Binop($1, Mult, $3) }
	| expr DIVIDE expr { Binop($1, Div, $3) }
	| expr EQ     expr { Binop($1, Equal, $3) }
	| expr NEQ    expr { Binop($1, Neq,   $3) }
	| expr LT     expr { Binop($1, Less,  $3) }
	| expr LEQ    expr { Binop($1, Leq,   $3) }
	| expr GT     expr { Binop($1, Greater,  $3) }
	| expr GEQ    expr { Binop($1, Geq,   $3) }
	| expr CONS  expr  { Binop($1, Cons, $3) } 
	| LITERAL          { Literal($1) }
	| BOOLEAN_LITERAL  { BooleanLiteral($1) }
	| LPAREN expr RPAREN { $2 }
	| ID ASSIGN expr   { Assign($1, $3) }
	| ID LPAREN actuals_opt RPAREN { Call($1, $3) }
	| VISIT ID { Visit($2) }
	| UNVISIT ID { Unvisit($2) }
	| LBRACK node_ids RBRACK { List($2) }
	| expr INC { Unop(Inc, $1) }
	| expr DEC { Unop(Dec, $1) }
	| NOT expr { Unop(Not, $2) }

/* list of node ids */
node_ids: 
	| /* nothing */ { [] }
	| ID { [Id($1)] }
	| ID COMMA node_ids { Id($1)::$3 }

actuals_opt:
	/* nothing */ { [] }
	| actuals_list  { List.rev $1 }

actuals_list:
	| expr                    { [$1] }
	| actuals_list COMMA expr { $3 :: $1 }






  
